;---------------------------------------------------------------
; Floppy driver
;
; Based on colorForth (7/31/01 release), modified for compatibility with
; older FDC's like the one in my 386. I had to use DMA because apparently
; PIO is too slow to keep up with the drive. After a seek or recal, I test
; SR0 bit 5 to check for completion (Chuck's was testing bit 7). Port 3f2 is
; write-only on older FDC's, so I use a variable to keep track of whether
; the drive is already spinning. And, I had to give an explicit seek
; command; I've heard that newer FDC's do an "implied seek" when necessary.

command	db 0
	db 0	;drive/head
cyl	db 0
	db 0	;head
	db 1	;sector
	db 2	;512 bytes/sector
	db 18	;sectors/track
	db 0x1B	;gap
	db 0xFF	;data length (ignored unless bytes/sector = 0)

spinning db 0	;drive is already spinning & calibrated

recal:	mov al,7		;recalibrate
	mov ecx,2
	call cmd
	call ready
.a:	call sense
	js .a
	ret

spin:	mov cl,1Ch		;start motor & recalibrate drive
	call onoff
	jmp recal

flop:	mov [cyl],al
	dup
	test byte [spinning],10h	;is drive already spinning?
	  jz spin
	ret

ready:	mov dx,3f4h
.a:	  in al,dx
	  out 0e1h,al
	  shl al,1
	  jnc .a
	lea edx,[edx+1]		;increment edx w/o touching flags
	ret

transfer:
	mov cl,9		;ecx = command length
cmd:	mov edx,command
	mov [edx],al
cmd0:	push esi
	mov esi,edx
cmd1:	call ready		;dx=3f5
	jns .a			;read & discard any waiting data
	  in al,dx
	  jmp cmd1
.a:	lodsb			;send command
	  out dx,al
	  out 0e1h,al
	  loop cmd1
	pop esi
	ret

sense:	mov al,8		;cmd: sense interrupt status
	mov ecx,1
	call cmd
.a:	  call ready
	  jns .a
	in al,dx
	out 0e1h,al
	and al,al
	ret

seek:	call sense		;wait until seek finishes
	test al,20h
	jz seek
	ret

stop:	mov cl,0Ch		;turn off all motors
onoff:	dup
	mov al,cl
	mov dx,3f2h
	out dx,al
	out 0e1h,al
	mov [spinning],al
	drop
	ret

macro PIO {
;PIO version
read:	mov al,0Fh		;seek
	mov ecx,3
	call cmd
	call seek
	mov al,0e6h
	call transfer
	mov cx,18*2*512		;1 cylinder
.a:	  call ready
	  in al,dx
	  out 0e1h,al
	  stosb
	  loop .a
	ret
}

;DMA version
read:	mov al,0Fh		;seek
	mov ecx,3
	call cmd
	call seek
	mov al,46h		;read
	call dma
	mov al,0E6h
	call transfer
	call ready
	push esi		;copy data from dma buffer
	mov esi, 0x100000
	mov ecx,1200h
	rep movsd
	pop esi
	ret

write:	mov al,0Fh		;seek
	mov ecx,3
	call cmd
	call seek
	push edi		;copy data to dma buffer
	mov edi,0x100000
	mov ecx,1200h
	rep movsd
	pop edi
	mov al,4Ah		;write
	call dma
	mov al,0C5h
	call transfer
	call ready
	ret

; al=mode
dma:	out 0Bh,al
	out 0Ch,al		;reset flip-flop
	mov eax,0x100000		;address (24 bits)
	out 04h,al
	shr eax,8
	out 04h,al
	shr eax,8
	out 81h,al
	mov al,0FFh		;length = sectors*512-1
	out 05h,al
	mov al,18*2-1
	out 05h,al
	mov al,2		;enable channel 2
	out 0Ah,al
	ret

;---------------------------------------------------------------
;	( a cyl -- a+ cyl+1 )
code 'readf', readf
        call flop
	push edi
	mov edi,[esi+4]
	call read
	pop edi
readf1:	drop
	inc eax
	add [esi],dword 4800h
	call stop
	ret

;	( a cyl -- a+ cyl+1 )
code 'writef', writef
  	call flop
	push esi
	mov esi,[esi+4]
	call write
	pop esi
	jmp readf1

;---------------------------------------------------------------
;	( a cyl n -- a+ cyl+n )
code 'reads', reads
	upop ecx
.a:	push ecx
	call readf
	pop ecx
	loop .a
	ret

;	( a cyl n -- a+ cyl+n )
code 'writes', writes
	upop ecx
.a:	push ecx
	call writef
	pop ecx
	loop .a
	ret

